# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Thomas Quintana <quintana.thomas@gmail.com>

from freepy.lib.fsm import *
from unittest import TestCase, expectedFailure

class AbstractLightBulb(FiniteStateMachine):
  # Initial state.
  initial_state = 'off'

  # Possible transitions.
  transitions = [
    ('off', 'on'),
    ('on', 'off'),
    ('off', 'broken'),
    ('on', 'broken')
  ]

class LightBulb(AbstractLightBulb):
  # Handle incoming events.
  def on_message(self, message):
    if message == 'turn on':
      self.transition(to = 'on', event = message)
    elif message == 'turn off':
      self.transition(to = 'off', event = message)
    elif message == 'break':
      self.transition(to = 'broken', event = message)

  @Guard(state = 'on')
  def check_electricity(self):
    if hasattr(self, 'electricity'):
      return self.electricity
    return False

  @Action(state = 'off')
  def turn_off(self, message):
    self.indicator = 'dim'

  @Action(state = 'on')
  def turn_on(self, message):
    self.indicator = 'lit'

  @Action(state = 'broken')
  def smash(self, message):
    self.indicator = 'broken'

class LightBulbWithMultipleGuards(AbstractLightBulb):
  @Guard(state = 'on')
  def check_electricity(self):
    pass

  @Guard(state = 'on')
  def check_socket(self):
    pass

class LightBulbWithMultipleOnEnterActions(AbstractLightBulb):
  @Action(state = 'on')
  def turn_on(self, message):
    pass

  @Action(state = 'on')
  def turn_on_again(self, message):
    pass

class LightBulbWithMultipleOnExitActions(AbstractLightBulb):
  @Action(state = 'on', on_enter = False, on_exit = True)
  def turn_on(self, message):
    pass

  @Action(state = 'on', on_enter = False, on_exit = True)
  def turn_on_again(self, message):
    pass

class LightBulbWithBadActionState(AbstractLightBulb):
  @Action(state = 'bogus')
  def foobar(self, message):
    pass

class LightBulbWithBadGuardState(AbstractLightBulb):
  @Guard(state = 'bogus')
  def foobar(self):
    pass

class LightBulbTests(TestCase):
  def test_success_scenario(self):
    light_bulb = LightBulb()
    light_bulb.electricity = True
    self.assertTrue(light_bulb.state() == 'off')
    light_bulb.on_message('turn on')
    self.assertTrue(light_bulb.state() == 'on')
    self.assertTrue(light_bulb.indicator == 'lit')
    light_bulb.on_message('turn off')
    self.assertTrue(light_bulb.state() == 'off')
    self.assertTrue(light_bulb.indicator == 'dim')

  def test_invalid_transition(self):
    light_bulb = LightBulb()
    light_bulb.electricity = True
    self.assertTrue(light_bulb.state() == 'off')
    self.assertRaises(FiniteStateMachineError, light_bulb.on_message, 'turn off')
    self.assertFalse(hasattr(light_bulb, 'indicator'))

  def test_terminal_state(self):
    light_bulb = LightBulb()
    light_bulb.electricity = True
    self.assertTrue(light_bulb.state() == 'off')
    light_bulb.on_message('turn on')
    self.assertTrue(light_bulb.state() == 'on')
    self.assertTrue(light_bulb.indicator == 'lit')
    light_bulb.on_message('turn off')
    self.assertTrue(light_bulb.state() == 'off')
    self.assertTrue(light_bulb.indicator == 'dim')
    light_bulb.on_message('break')
    self.assertTrue(light_bulb.state() == 'broken')
    self.assertTrue(light_bulb.indicator == 'broken')
    self.assertRaises(FiniteStateMachineError, light_bulb.on_message, 'turn on')
    self.assertTrue(light_bulb.state() == 'broken')
    self.assertTrue(light_bulb.indicator == 'broken')

  def test_guard_denial(self):
    light_bulb = LightBulb()
    self.assertTrue(light_bulb.state() == 'off')
    self.assertRaises(FiniteStateMachineError, light_bulb.on_message, 'turn on')
    self.assertTrue(light_bulb.state() == 'off')
    self.assertFalse(hasattr(light_bulb, 'indicator'))

  def test_multiple_guards(self):
    light_bulb = None
    self.assertRaises(FiniteStateMachineError, LightBulbWithMultipleGuards)
    self.assertTrue(light_bulb == None)

  def test_multiple_on_enter_actions(self):
    light_bulb = None
    self.assertRaises(FiniteStateMachineError, LightBulbWithMultipleOnEnterActions)
    self.assertTrue(light_bulb == None)

  def test_multiple_on_exit_actions(self):
    light_bulb = None
    self.assertRaises(FiniteStateMachineError, LightBulbWithMultipleOnExitActions)
    self.assertTrue(light_bulb == None)

  def test_bad_action_state(self):
    light_bulb = None
    self.assertRaises(FiniteStateMachineError, LightBulbWithBadActionState)
    self.assertTrue(light_bulb == None)

  def test_bad_guard_state(self):
    light_bulb = None
    self.assertRaises(FiniteStateMachineError, LightBulbWithBadGuardState)
    self.assertTrue(light_bulb == None)


# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Thomas Quintana <quintana.thomas@gmail.com>

from os import SEEK_END
from twisted.internet.protocol import Protocol, ReconnectingClientFactory

# Import the proper StringIO implementation.
try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

import logging

class IEventSocketClientObserver(object):
  def on_event(self, event):
    pass

  def on_start(self, client):
    pass

  def on_stop(self):
    pass

class Event(object):
  def __init__(self, headers, body = None):
    self.__headers__ = headers
    self.__body__ = body

  def get_body(self):
    return self.__body__

  def get_header(self, name):
    return self.__headers__.get(name)

  def get_headers(self):
    return self.__headers__

class EventSocketClient(Protocol):
  def __init__(self, observer):
    self.__logger__ = logging.getLogger('freepy.lib.esl.eventsocketclient')
    # Data buffer.
    self.__buffer__ = None
    # Client state.
    self.__host__ = None
    self.__peer__ = None
    # Observer for incoming events.
    if isinstance(observer, IEventSocketClientObserver):
      self.__observer__ = observer
    else:
      raise TypeError('The observer must extend the \
      IEventSocketClientObserver interface.')

  def __parse__(self):
    # Make sure we have enough data to process the event.
    buffer_contents = self.__buffer__.getvalue()
    if len(buffer_contents) == 0 or not buffer_contents.find('\n\n'):
      return None
    else:
      # Parse the event for processing.
      self.__buffer__.seek(0)
      body = None
      headers = self.__parse_headers__()
      length = headers.get('Content-Length')
      if length:
        length = int(length)
        # Remove the Content-Length header.
        del headers['Content-Length']
        # Make sure we have enough data to process the body.
        offset = self.__buffer__.tell()
        self.__buffer__.seek(0, SEEK_END)
        end = self.__buffer__.tell()
        remaining = end - offset
        # Handle the event body.
        if length <= remaining:
          self.__buffer__.seek(offset)
          type = headers.get('Content-Type')
          if type and type == 'text/event-plain':
            headers.update(self.__parse_headers__())
            length = headers.get('Content-Length')
            if length:
              length = int(length)
              del headers['Content-Length']
              body = self.__buffer__.read(length)
          else:
            body = self.__buffer__.read(length)
        else:
          return None
      # Reclaim resources.
      offset = self.__buffer__.tell()
      self.__buffer__.seek(0, SEEK_END)
      end = self.__buffer__.tell()
      remaining = end - offset
      if remaining == 0:
        self.__buffer__.seek(0)
        self.__buffer__.truncate(0)
      else:
        self.__buffer__.seek(offset)
        data = self.__buffer__.read(remaining)
        self.__buffer__.seek(0)
        self.__buffer__.write(data)
        self.__buffer__.truncate(remaining)
      return Event(headers, body)

  def __parse_headers__(self):
    headers = dict()
    while True:
      line = self.__parse_line__()
      if line == '':
        break
      else:
        tokens = line.split(':', 1)
        name = tokens[0].strip()
        value = tokens[1].strip()
        headers.update({name: value})
    return headers

  def __parse_line__(self, stride = 64):
    line = list()
    while True:
      chunk = self.__buffer__.read(stride)
      end = chunk.find('\n')
      if end == -1:
        line.append(chunk)
      else:
        line.append(chunk[:end])
        offset = self.__buffer__.tell()
        left_over = len(chunk[end + 1:])
        self.__buffer__.seek(offset - left_over)
        break
      if len(chunk) < stride:
        break
    return ''.join(line)

  def connectionLost(self, reason):
    self.__logger__.critical('A connection to the FreeSWITCH instance located @ %s:%i \
    has been lost due to the following reason.\n%s', self.__peer__.host, 
    self.__peer__.port, reason)
    self.__observer__.on_stop()
    if self.__buffer__:
      self.__buffer__.close()
    self.__buffer__ = None
    self.__host__ = None
    self.__peer__ = None

  def connectionMade(self):
    self.__buffer__ = StringIO()
    self.__host__ = self.transport.getHost()
    self.__peer__ = self.transport.getPeer()
    self.__observer__.on_start(self)

  def dataReceived(self, data):
    if self.__logger__.isEnabledFor(logging.DEBUG):
      self.__logger__.debug('The following message was received from %s:%i.\n%s',
        self.__peer__.host, self.__peer__.port, data)
    self.__buffer__.write(data)
    while True:
      event = self.__parse__()
      if event:
        self.__observer__.on_event(event)
      else:
        break

  def send(self, command):
    serialized_command = str(command)
    if self.__logger__.isEnabledFor(logging.DEBUG):
      self.__logger__.debug('The following message will be sent to %s:%i.\n%s',
        self.__peer__.host, self.__peer__.port, serialized_command)
    self.transport.write(serialized_command)

class EventSocketClientFactory(ReconnectingClientFactory):
  def __init__(self, observer):
    self.__logger__ = logging.getLogger('freepy.lib.esl.eventsocketclientfactory')
    self.__observer__ = observer

  def buildProtocol(self, addr):
    if self.__logger__.isEnabledFor(logging.INFO):
      self.__logger__.info('Connected to the FreeSWITCH instance located @ %s:%i.',
        addr.host, addr.port)
    self.resetDelay()
    return EventSocketClient(self.__observer__)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Thomas Quintana <quintana.thomas@gmail.com>

from llist import dllist
from pykka import ThreadingActor
from threading import Thread

import logging, time

class ServiceRequest(object):
  pass

class ReceiveTimeoutCommand(ServiceRequest):
  def __init__(self, sender, timeout, recurring = False):
    self.__sender__ = sender
    self.__timeout__ = timeout
    self.__recurring__ = recurring

  def get_sender(self):
    return self.__sender__

  def get_timeout(self):
    return self.__timeout__

  def is_recurring(self):
    return self.__recurring__

class StopTimeoutCommand(ServiceRequest):
  def __init__(self, sender):
    self.__sender__ = sender

  def get_sender(self):
    return self.__sender__

class ClockEvent(object):
  pass

class TimeoutEvent(object):
  pass

class MonotonicClock(Thread):
  def __init__(self, *args, **kwargs):
    super(MonotonicClock, self).__init__(group = None)
    self.__actor__ = args[0]
    self.__interval__ = args[1]
    self.__running__ = True
    # Singleton instance of ClockEvent.
    self.__event__ = ClockEvent()

  def run(self):
    while self.__running__:
      time.sleep(self.__interval__)
      if self.__running__:
        self.__actor__.tell({'content': self.__event__})

  def stop(self):
    self.__running__ = False

class TimerService(ThreadingActor):
  '''
  The timer service uses the timing wheel algorithm borrowing from the
  approach used in the Linux kernel. Please refer to the email thread
  by Ingo Molnar @ https://lkml.org/lkml/2005/10/19/46.
  '''
  TICK_SIZE  = 0.1             # Tick every 100 milliseconds.

  def __init__(self, *args, **kwargs):
    super(TimerService, self).__init__(*args, **kwargs)
    # Initialize the timing wheels. The finest possible
    self.__logger__ = logging.getLogger('freepy.lib.services.TimerService')
    # granularity is 100ms.
    self.__timer_vector1__ = self.__create_vector__(256)
    self.__timer_vector2__ = self.__create_vector__(256)
    self.__timer_vector3__ = self.__create_vector__(256)
    self.__timer_vector4__ = self.__create_vector__(256)
    # Initialize the timer vector indices.
    self.__timer_vector2_index__  = 0
    self.__timer_vector3_index__  = 0
    self.__timer_vector4_index__  = 0
    # Initialize the tick counter.
    self.__current_tick__ = 0
    # Initialize the actor lookup table for O(1) timer removal.
    self.__actor_lookup_table__ = dict()
    # Singleton instance of the timeout event.
    self.__timeout__ = TimeoutEvent()
    # Monotonic clock.
    self.__clock__ = None

  def __cascade_vector__(self, vector, elapsed):
    '''
    Cascades all the timers inside a vector to a lower bucket.

    Arguments: vector  - The vector to cascade.
               elapsed - The amount of time elapsed in milliseconds.
    '''
    for index in range(1, len(vector)):
      bucket = vector[index]
      previous_bucket = vector[index - 1]
      while len(bucket) > 0:
        timer = bucket.popleft()
        expires = timer.get_expires() - elapsed
        timer.set_expires(expires)
        node = previous_bucket.append(timer)
        self.__update_lookup_table__(previous_bucket, node)

  def __cascade_vector_2__(self):
    '''
    Cascades timers from vector 2 into vector 1.
    '''
    tick = self.__current_tick__
    timers = self.__timer_vector2__[0]
    for timer in timers:
      expires = timer.get_expires() - 25600
      timer.set_expires(expires)
      self.__vector1_insert__(timer)
    timers.clear()
    self.__cascade_vector__(self.__timer_vector2__, 25600)
    index = self.__timer_vector2_index__
    index = (index + 1) % 256
    self.__timer_vector2_index__ = index
    if self.__timer_vector2_index__ == 0:
      self.__cascade_vector_3__()

  def __cascade_vector_3__(self):
    '''
    Cascades timers from vector 3 into vector 2.
    '''
    tick = self.__current_tick__
    timers = self.__timer_vector3__[0]
    for timer in timers:
      expires = timer.get_expires() - 6553600
      timer.set_expires(expires)
      self.__vector2_insert__(timer)
    timers.clear()
    self.__cascade_vector__(self.__timer_vector3__, 6553600)
    index = self.__timer_vector3_index__
    index = (index + 1) % 256
    self.__timer_vector3_index__ = index
    if self.__timer_vector3_index__ == 0:
      self.__cascade_vector_4__()

  def __cascade_vector_4__(self):
    '''
    Cascades timers from vector 4 into vector 3.
    '''
    tick = self.__current_tick__
    timers = self.__timer_vector4__[0]
    for timer in timers:
      expires = timer.get_expires() - 1677721600
      timer.set_expires(expires)
      self.__vector3_insert__(timer)
    timers.clear()
    self.__cascade_vector__(self.__timer_vector4__, 1677721600)
    index = self.__timer_vector4_index__
    index = (index + 1) % 256
    self.__timer_vector4_index__ = index

  def __create_vector__(self, size):
    '''
    Creates a new vector and initializes it to a specified size.

    Arguments: size - The size of the vector.
    '''
    vector = list()
    for counter in range(size):
      vector.append(dllist())
    return vector

  def __round__(self, timeout):
    '''
    Rounds a timeout to the nearest multiple of the tick size.

    Arguments: timeout - The timeout to be rounded.
    '''
    remainder = timeout % 100
    if remainder == 0:
      return timeout
    elif remainder <= 49:
      return  timeout - remainder
    else:
      return timeout + 100 - remainder

  def __schedule__(self, timer):
    '''
    Schedules a timer for expiration.

    Arguments: timer - The timer to be shceduled.
    '''
    tick = self.__current_tick__
    timeout = timer.get_timeout()
    expires = (tick % 256) * 100 + self.__round__(timeout)
    timer.set_expires(expires)
    if expires <= 25600:
      self.__vector1_insert__(timer)
    elif expires <= 6553600:
      self.__vector2_insert__(timer)
    elif expires <= 1677721600:
      self.__vector3_insert__(timer)
    elif expires <= 429496729600:
      self.__vector4_insert__(timer)

  def __tick__(self):
    '''
    Excutes one clock tick.
    '''
    tick = self.__current_tick__
    timers = self.__timer_vector1__[tick % 256]
    recurring = list()
    while len(timers) > 0:
      timer = timers.popleft()
      timer.get_observer().tell({'content': self.__timeout__})
      if timer.is_recurring():
        recurring.append(timer)
      else:
        lookup_table = self.__actor_lookup_table__
        urn = timer.get_observer.actor_urn
        location = lookup_table.get(urn)
        if location:
          del lookup_table[location]
    for timer in recurring:
      self.__schedule__(timer)
    self.__current_tick__ = tick + 1
    if self.__current_tick__ % 256 == 0:
      self.__cascade_vector_2__()

  def __unschedule__(self, command):
    '''
    Unschedules a timer that has previously been scheduled for expiration.

    Arguments: command - The StopTimeoutCommand.
    '''
    urn = command.get_sender().actor_urn
    location = self.__actor_lookup_table__.get(urn)
    if location:
      del self.__actor_lookup_table__[urn]
      vector = location.get('vector')
      node = location.get('node')
      vector.remove(node)

  def __update_lookup_table__(self, vector, node):
    '''
    Updates a lookup table used for O(1) timer removal.
    '''
    urn = node.value.get_observer().actor_urn
    location = {
      'vector': vector,
      'node': node
    }
    self.__actor_lookup_table__.update({urn: location})

  def __vector1_insert__(self, timer):
    '''
    Inserts a timer into vector 1.

    Arguments: timer - The timer to be inserted.
    '''
    vector = self.__timer_vector1__
    bucket = timer.get_expires() / 100 - 1
    node = vector[bucket].append(timer)
    self.__update_lookup_table__(vector[bucket], node)

  def __vector2_insert__(self, timer):
    '''
    Inserts a timer into vector 2.

    Arguments: timer - The timer to be inserted.
    '''
    vector = self.__timer_vector2__
    bucket = timer.get_expires() / 25600 - 1
    node = vector[bucket].append(timer)
    self.__update_lookup_table__(vector[bucket], node)

  def __vector3_insert__(self, timer):
    '''
    Inserts a timer into vector 3.

    Arguments: timer - The timer to be inserted.
    '''
    vector = self.__timer_vector3__
    bucket = timer.get_expires() / 6553600 - 1
    node = vector[bucket].append(timer)
    self.__update_lookup_table__(vector[bucket], node)

  def __vector4_insert__(self, timer):
    '''
    Inserts a timer into vector 4.

    Arguments: timer - The timer to be inserted.
    '''
    vector = self.__timer_vector4__
    bucket = timer.get_expires() / 1677721600 - 1
    node = vector[bucket].append(timer)
    self.__update_lookup_table__(vector[bucket], node)

  def on_failure(self, exception_type, exception_value, traceback):
    '''
    Logs exceptions for the TimerService.
    '''
    self.__logger__.error(exception_value)

  def on_receive(self, message):
    '''
    Handles incoming messages.

    Arguments: message - The message to be processed.
    '''
    # This is necessary because all Pykka messages
    # must be of type dict.
    message = message.get('content')
    if not message:
      return
    # Handle the message.
    if isinstance(message, ReceiveTimeoutCommand):
      uuid = message.get_sender().actor_urn
      if not self.__actor_lookup_table__.has_key(uuid):
        timeout = message.get_timeout()
        observer = message.get_sender()
        recurring = message.is_recurring()
        timer = TimerService.Timer(observer, timeout, recurring)
        self.__schedule__(timer)
      else:
        self.__logger__.warning('Actor %s is requesting too many simultaneous timers.'
          % uuid)
    elif isinstance(message, StopTimeoutCommand):
      self.__unschedule__(message)
    elif isinstance(message, ClockEvent):
      self.__tick__()

  def on_start(self):
    '''
    Initialized the TimerService.
    '''
    self.__clock__ = MonotonicClock(self.actor_ref, TimerService.TICK_SIZE)
    self.__clock__.start()

  def on_stop(self):
    '''
    Cleans up after TimerService.
    '''
    self.__clock__.stop()

  class Timer(object):
    def __init__(self, observer, timeout, recurring = False):
      self.__observer__ = observer
      self.__recurring__ = recurring
      self.__timeout__ = timeout
      self.__expires__ = 0

    def get_expires(self):
      return self.__expires__

    def get_observer(self):
      return self.__observer__

    def get_timeout(self):
      return self.__timeout__

    def is_recurring(self):
      return self.__recurring__

    def set_expires(self, expires):
      self.__expires__ = expires



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Thomas Quintana <quintana.thomas@gmail.com>
#
# Cristian Groza <frontc18@gmail.com>

from freepy.lib.commands import *
from unittest import TestCase, expectedFailure

class BackgroundCommandTests(TestCase):
  def test_success_scenario(self):
    command = BackgroundCommand(object())

class UUIDCommandTests(TestCase):
  def test_failed_scenario_no_variables(self):
    self.assertRaises(ValueError, UUIDCommand, None, None)

class ACLCheckCommandTests(TestCase):
	def test_success_scenario(self):
		command = ACLCheckCommand(object(), ip = '192.168.1.1', list_name = 'lan')
		desired_output = 'bgapi acl 192.168.1.1 lan\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_failed_scenario_no_variables(self):
		self.assertRaises(ValueError, ACLCheckCommand, object(), ip = None, list_name = None)

	def test_failed_scenario_no_variable_ip(self):
		self.assertRaises(ValueError, ACLCheckCommand, object(), ip = None, list_name = 'lan')

	def test_failed_scenario_no_variable_list_name(self):
		self.assertRaises(ValueError, ACLCheckCommand, object(), ip = '192.168.1.1', list_name = None)

class CheckUserGroupCommandTests(TestCase):
	def test_success_scenario(self):
		command = CheckUserGroupCommand(object(), user = 'test', domain = 'domain_test.com', group_name = 'groupname_test')
		desired_output = 'bgapi in_group test@domain_test.com groupname_test\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class DialedExtensionHupAllCommandTests(TestCase):
	def test_success_scenario(self):
		command = DialedExtensionHupAllCommand(object(), clearing = '999-999-9999', extension = '493')
		desired_output = 'bgapi fsctl hupall 999-999-9999 dialed_ext 493\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)
    
class DisableVerboseEventsCommandTests(TestCase):
	def test_success_scenario(self):
		command = DisableVerboseEventsCommand(object())
		desired_output = 'bgapi fsctl verbose_events off\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class DomainExistsCommandTests(TestCase):
	def test_success_scenario(self):
		command = DomainExistsCommand(object(), domain = 'domain_test.com')
		desired_output = 'bgapi domain_exists domain_test.com\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class EnableVerboseEventsCommandTests(TestCase):
  def test_success_scenario(self):
    command = EnableVerboseEventsCommand(object())
    desired_output = 'bgapi fsctl verbose_events on\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

class GetDefaultDTMFDurationCommandTests(TestCase):
  def test_success_scenario(self):
    command = GetDefaultDTMFDurationCommand(object())
    desired_output = 'bgapi fsctl default_dtmf_duration 0\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)
    
class GetGlobalVariableCommandTests(TestCase):
  def test_success_scenario(self):
    command = GetGlobalVariableCommand(object(), name='testVariable')   	
    desired_output = 'bgapi global_getvar testVariable\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

  def test_failed_scenario_no_variable(self):
  	self.assertRaises(ValueError, GetGlobalVariableCommand, object(), name = None)

class GetMaxSessionsCommandTests(TestCase):
  def test_success_scenario(self):
    command = GetMaxSessionsCommand(object())   	
    desired_output = 'bgapi fsctl max_sessions\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

class GetMaximumDTMFDurationCommandTests(TestCase):
  def test_success_scenario(self):
    command = GetMaximumDTMFDurationCommand(object())   	
    desired_output = 'bgapi fsctl max_dtmf_duration 0\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

class GetMinimumDTMFDurationCommandTests(TestCase):
  def test_success_scenario(self):
    command = GetMinimumDTMFDurationCommand(object())   	
    desired_output = 'bgapi fsctl min_dtmf_duration 0\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

class GetSessionsPerSecondCommandTests(TestCase):
  def test_success_scenario(self):
    command = GetSessionsPerSecondCommand(object())   	
    desired_output = 'bgapi fsctl last_sps\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

class GetGroupCallBridgeStringCommandTests(TestCase):
  def test_success_scenario_value_F(self):
    command = GetGroupCallBridgeStringCommand(object(), group = 'groupname_test', domain = 'domain_test.com', option = '+F')   	
    desired_output = 'bgapi group_call groupname_test@domain_test.com+F\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

  def test_failed_scenario_no_value_option(self):
  	self.assertRaises(ValueError, GetGroupCallBridgeStringCommand, object(), group = 'groupname_test', domain = 'domain_test.com', option = 'None')

  def test_success_scenario_value_A(self):
    command = GetGroupCallBridgeStringCommand(object(), group = 'groupname_test', domain = 'domain_test.com', option = '+A')   	
    desired_output = 'bgapi group_call groupname_test@domain_test.com+A\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)
   
  def test_success_scenario_value_F(self):
    command = GetGroupCallBridgeStringCommand(object(), group = 'groupname_test', domain = 'domain_test.com', option = '+E')   	
    desired_output = 'bgapi group_call groupname_test@domain_test.com+E\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

  def test_failed_scenario_bad_variable(self):
    self.assertRaises(ValueError, GetGroupCallBridgeStringCommand, object(), group = 'groupname_test', domain = 'domain_test.com', option = '+G')
 
class HupAllCommandTests(TestCase):
  def test_success_scenario(self):
    command = HupAllCommand(object(), cause = 'test_cause', var_name = 'test_var_name', var_value = 'test_var_value')   	
    desired_output = 'bgapi hupall test_cause test_var_name test_var_value\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)
 
class LoadModuleCommandTests(TestCase):
  def test_success_scenario(self):
    command = LoadModuleCommand(object(), name = "testModule")
    desired_output = 'bgapi load testModule\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)
     
class OriginateCommandTests(TestCase):
  def test_success_scenario(self):
    command = OriginateCommand(object(), url = "http://test.com", extension = None, app_name = 'appTestme', app_args = ['arg1','arg2'], options = ['opt1','opt2'])
    desired_output = 'bgapi originate {opt1,opt2}http://test.com &appTestme(arg1 arg2)\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)
    
class PauseSessionCreationCommandTests(TestCase):
  def test_success_scenario(self):
    command = PauseSessionCreationCommand(object(), direction = "testDirection")
    desired_output = 'bgapi fsctl pause testDirection\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

class ReclaimMemoryCommandTests(TestCase):
  def test_success_scenario(self):
    command = ReclaimMemoryCommand(object())
    desired_output = 'bgapi fsctl reclaim_mem\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

class ResumeSessionCreationCommandTests(TestCase):
  def test_success_scenario_passvariable(self):
    command = ResumeSessionCreationCommand(object(), direction = "testDirection")
    desired_output = 'bgapi fsctl resume testDirection\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)
   	
  def test_success_scenario_novariable(self):
    command = ResumeSessionCreationCommand(object())
    desired_output = 'bgapi fsctl resume\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

class SetDefaultDTMFDurationCommandTests(TestCase):
  def test_success_scenario(self):
    command = SetDefaultDTMFDurationCommand(object(), duration = 500)
    desired_output = 'bgapi fsctl default_dtmf_duration 500\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

class SetGlobalVariableCommandTests(TestCase):
  def test_success_scenario(self):
    command = SetGlobalVariableCommand(object(), name = 'testName', value = 'testValue')
    desired_output = 'bgapi global_setvar testName=testValue\nJob-UUID: %s\n\n'  % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

class SetMaximumDTMFDurationCommandTests(TestCase):
	def test_success_scenario(self):
		command = SetMaximumDTMFDurationCommand(object(), duration = 500)
		desired_output = 'bgapi fsctl max_dtmf_duration 500\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class SetMinimumDTMFDurationCommandTests(TestCase):
	def test_success_scenario(self):
		command = SetMinimumDTMFDurationCommand(object(), duration = 500)
		desired_output = 'bgapi fsctl min_dtmf_duration 500\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class SetSessionsPerSecondCommandTests(TestCase):
	def test_success_scenario(self):
		command = SetSessionsPerSecondCommand(object(), sessions_per_second = 15)
		desired_output = 'bgapi fsctl sps 15\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class ShutdownCommandTests(TestCase):
	def test_success_scenario_value_elegant(self):
		command = ShutdownCommand(object(), option = 'elegant')
		desired_output = 'bgapi fsctl shutdown elegant\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_success_scenario_value_asap(self):
		command = ShutdownCommand(object(), option = 'asap')
		desired_output = 'bgapi fsctl shutdown asap\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_success_scenario_value_restart(self):
		command = ShutdownCommand(object(), option = 'restart')
		desired_output = 'bgapi fsctl shutdown restart\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_success_scenario_value_restart(self):
		command = ShutdownCommand(object(), option = 'cancel')
		desired_output = 'bgapi fsctl shutdown cancel\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_failed_scenario_with_no_value(self):
		self.assertRaises(ValueError, ShutdownCommand, object(), option = 'n/a')

class StatusCommandTests(TestCase):
	def test_success_scenario(self):
		command = StatusCommand(object())
		desired_output = 'bgapi status\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)
    
class SyncClockCommandTests(TestCase):
	def test_success_scenario(self):
		command = SyncClockCommand(object())
		desired_output = 'bgapi fsctl sync_clock\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class SyncClockWhenIdleCommandTests(TestCase):
	def test_success_scenario(self):
		command = SyncClockWhenIdleCommand(object())
		desired_output = 'bgapi fsctl sync_clock_when_idle\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class UnloadModuleCommandTests(TestCase):
	def test_success_scenario_passvariables(self):
		command = UnloadModuleCommand(object(), name = 'testName', force = 'true')
		desired_output = 'bgapi unload -f testName\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)
   
	def test_success_scenario_passvariable(self):
		command = UnloadModuleCommand(object(), name = 'testName')
		desired_output = 'bgapi unload testName\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class AnswerCommandTests(TestCase):
	def test_success_scenario(self):
		command = AnswerCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_answer 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class BreakCommandTests(TestCase):
	def test_success_scenario(self):
		command = BreakCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_break 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class BridgeCommandTests(TestCase):
	def test_success_scenario(self):
		command = BridgeCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', other_uuid = '21516b8e-5a0b-485a-9e53-933e42947666')
		desired_output = 'bgapi uuid_bridge 21516b8e-5a0b-485a-9e53-933e42947079 21516b8e-5a0b-485a-9e53-933e42947666\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_failed_scenario_no_variable_other_uuid(self):
		self.assertRaises(ValueError, BridgeCommand, object(), '21516b8e-5a0b-485a-9e53-933e42947079', other_uuid = None)

class BroadcastCommandTests(TestCase):
	def test_success_scenario_withPath(self):
		command = BroadcastCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', leg = 'aleg', path = 'testPath', app_name = None, app_args = 'app_args')
		desired_output = 'bgapi uuid_broadcast 21516b8e-5a0b-485a-9e53-933e42947079 testPath aleg\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_success_scenario_withAppName(self):
		command = BroadcastCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', leg = 'aleg', path = None, app_name = 'testApp', app_args = 'app_args')
		desired_output = 'bgapi uuid_broadcast 21516b8e-5a0b-485a-9e53-933e42947079 testApp::app_args aleg\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_success_scenario_variable_leg_value_aleg(self):
		command = BroadcastCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', leg = 'aleg', path = 'testPath', app_name = None, app_args = 'app_args')
		desired_output = 'bgapi uuid_broadcast 21516b8e-5a0b-485a-9e53-933e42947079 testPath aleg\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_success_scenario_variable_leg_value_bleg(self):
		command = BroadcastCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', leg = 'bleg', path = 'testPath', app_name = None, app_args = 'app_args')
		desired_output = 'bgapi uuid_broadcast 21516b8e-5a0b-485a-9e53-933e42947079 testPath bleg\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_success_scenario_variable_leg_value_both(self):
		command = BroadcastCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', leg = 'both', path = 'testPath', app_name = None, app_args = 'app_args')
		desired_output = 'bgapi uuid_broadcast 21516b8e-5a0b-485a-9e53-933e42947079 testPath both\nJob-UUID: %s\n\n'  % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_failed_scenario_no_variable_leg(self):
		self.assertRaises(ValueError, BroadcastCommand, object(), '21516b8e-5a0b-485a-9e53-933e42947079', leg = None, path = 'testPath', app_name = None, app_args = 'app_args')

	def test_failed_scenario_runtime(self):
		self.assertRaises(RuntimeError, BroadcastCommand, object(), '21516b8e-5a0b-485a-9e53-933e42947079', leg = 'bleg', path = 'testPath', app_name = 'testApp', app_args = 'app_args')

class ChatCommandTests(TestCase):
	def test_success_scenario(self):
		command = ChatCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', text = None)
		desired_output = 'bgapi uuid_chat 21516b8e-5a0b-485a-9e53-933e42947079 None\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)    

class DeflectCommandTests(TestCase):
	def test_success_scenario(self):
		command = DeflectCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', url = 'testURL')
		desired_output = 'bgapi uuid_deflect 21516b8e-5a0b-485a-9e53-933e42947079 testURL\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)    

class DisableMediaCommandTests(TestCase):
	def test_success_scenario(self):
		command = DisableMediaCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_media off 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)    

class DisplayCommandTests(TestCase):
	def test_success_scenario(self):
		command = DisplayCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', display = 'MessageBox.Show()')
		desired_output = 'bgapi uuid_display 21516b8e-5a0b-485a-9e53-933e42947079 MessageBox.Show()\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)    

class DualTransferCommandTests(TestCase):
	def test_success_scenario(self):
		command = DualTransferCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', extension_a = '0', extension_b = '1', dialplan_a = 'plana', dialplan_b = 'planb', context_a = 'contxa', context_b = 'contxb')
		desired_output = 'bgapi uuid_dual_transfer 21516b8e-5a0b-485a-9e53-933e42947079 0/plana/contxa 1/planb/contxb\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_failed_scenario_without_variable(self):
		self.assertRaises(RuntimeError, DualTransferCommand, object(), '21516b8e-5a0b-485a-9e53-933e42947079', extension_a = None, extension_b = None)    

class DumpCommandTests(TestCase):
	def test_success_scenario(self):
		command = DumpCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', format = None)
		desired_output = 'bgapi uuid_dump 21516b8e-5a0b-485a-9e53-933e42947079 None\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)    

class EarlyOkayCommandTests(TestCase):
	def test_success_scenario(self):
		command = EarlyOkayCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_early_ok 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class EnableMediaCommandTests(TestCase):
	def test_success_scenario(self):
		command = EnableMediaCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_media 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class EnableSessionHeartbeatCommandTests(TestCase):
	def test_success_scenario_with_variable(self):
		command = EnableSessionHeartbeatCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', start_time = 60)
		desired_output = 'bgapi uuid_session_heartbeat 21516b8e-5a0b-485a-9e53-933e42947079 sched 60\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_success_scenario_without_variable(self):
		command = EnableSessionHeartbeatCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_session_heartbeat 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class FileManagerCommandTests(TestCase):
  def test_success_scenario_with_variable_speed(self):
    command = FileManagerCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', command = 'speed')
    desired_output = 'bgapi uuid_fileman 21516b8e-5a0b-485a-9e53-933e42947079 speed\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

  def test_success_scenario_with_variable_pause(self):
    command = FileManagerCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', command = 'pause')
    desired_output = 'bgapi uuid_fileman 21516b8e-5a0b-485a-9e53-933e42947079 pause\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

  def test_success_scenario_with_variable_truncate(self):
    command = FileManagerCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', command = 'truncate')
    desired_output = 'bgapi uuid_fileman 21516b8e-5a0b-485a-9e53-933e42947079 truncate\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

  def test_success_scenario_with_variable_volume(self):
    command = FileManagerCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', command = 'volume')
    desired_output = 'bgapi uuid_fileman 21516b8e-5a0b-485a-9e53-933e42947079 volume\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

  def test_success_scenario_with_variable_restart(self):
    command = FileManagerCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', command = 'restart')
    desired_output = 'bgapi uuid_fileman 21516b8e-5a0b-485a-9e53-933e42947079 restart\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

  def test_success_scenario_with_variable_seek(self):
    command = FileManagerCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', command = 'seek')
    desired_output = 'bgapi uuid_fileman 21516b8e-5a0b-485a-9e53-933e42947079 seek\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

  def test_failed_scenario_without_variable(self):
    self.assertRaises(ValueError, FileManagerCommand, object(), '21516b8e-5a0b-485a-9e53-933e42947079', command = None)

class FlushDTMFCommandTests(TestCase):
	def test_success_scenario(self):
		command = FlushDTMFCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_flush_dtmf 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class GetAudioLevelCommandTests(TestCase):
	def test_success_scenario(self):
		command = GetAudioLevelCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_audio 21516b8e-5a0b-485a-9e53-933e42947079 start read level\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class GetBugListCommandTests(TestCase):
	def test_success_scenario(self):
		command = GetBugListCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_buglist 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class GetVariableCommandTests(TestCase):
	def test_success_scenario(self):
		command = GetVariableCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', name = 'testName')
		desired_output = 'bgapi uuid_getvar 21516b8e-5a0b-485a-9e53-933e42947079 testName\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_failed_scenario_without_variable(self):
		self.assertRaises(ValueError, GetVariableCommand, object(), '21516b8e-5a0b-485a-9e53-933e42947079', name = None)

class HoldCommandTests(TestCase):
	def test_success_scenario(self):
		command = HoldCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_hold 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class KillCommandTests(TestCase):
	def test_success_scenario_with_variable(self):
		command = KillCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', cause = 'fake cause')
		desired_output = 'bgapi uuid_kill 21516b8e-5a0b-485a-9e53-933e42947079 fake cause\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_success_scenario_without_variable(self):
		command = KillCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_kill 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class LimitCommandTests(TestCase):
	def test_success_scenario(self):
		command = LimitCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', backend = 'backend', realm = 'realm', resource='resource', max_calls = 2, interval = 20, number = 15, dialplan = 'c', context='unknown')
		desired_output = 'bgapi uuid_limit 21516b8e-5a0b-485a-9e53-933e42947079 backend realm resource 2/20 15 c unknown\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)
   
class ParkCommandTests(TestCase):
	def test_success_scenario(self):
		command = ParkCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_park 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class PauseCommandTests(TestCase):
	def test_success_scenario(self):
		command = PauseCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi pause 21516b8e-5a0b-485a-9e53-933e42947079 on\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class PreProcessCommandTests(TestCase):
	def test_success_scenario(self):
		command = PreProcessCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_preprocess 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class PreAnswerCommandTests(TestCase):
	def test_success_scenario(self):
		command = PreAnswerCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_preanswer 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class ReceiveDTMFCommandTests(TestCase):
  def test_success_scenario(self):
    command = ReceiveDTMFCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
    desired_output = 'bgapi uuid_recv_dtmf 21516b8e-5a0b-485a-9e53-933e42947079 None\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

class RenegotiateMediaCommandTests(TestCase):
	def test_success_scenario(self):
		command = RenegotiateMediaCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', codec = 'test_codex94')
		desired_output = 'bgapi uuid_media_reneg 21516b8e-5a0b-485a-9e53-933e42947079 =test_codex94\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class SendDTMFCommandTests(TestCase):
	def test_success_scenario(self):
		command = SendDTMFCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_send_dtmf 21516b8e-5a0b-485a-9e53-933e42947079 None\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class SendInfoCommandTests(TestCase):
	def test_success_scenario(self):
		command = SendInfoCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
		desired_output = 'bgapi uuid_send_info 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class SetAudioLevelCommandTests(TestCase):
	def test_success_scenario_with_variable(self):
		command = SetAudioLevelCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', level = 3.3)
		desired_output = 'bgapi uuid_audio 21516b8e-5a0b-485a-9e53-933e42947079 start write level 3.300000\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

	def test_failed_scenario_with_invalid_variable_high(self):
		self.assertRaises(ValueError, SetAudioLevelCommand, object(), '21516b8e-5a0b-485a-9e53-933e42947079', level = 4.1)

	def test_failed_scenario_with_invalid_variable_low(self):
		self.assertRaises(ValueError, SetAudioLevelCommand, object(), '21516b8e-5a0b-485a-9e53-933e42947079', level = -4.1)

class SetMultipleVariableCommandTests(TestCase):
	def test_success_scenario(self):
		command = SetMultipleVariableCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', variables = dict([('49','55')]))
		desired_output = 'bgapi uuid_setvar_multi 21516b8e-5a0b-485a-9e53-933e42947079 4=9\nJob-UUID: %s\n\n' % command.__job_uuid__
		self.assertTrue(str(command) == desired_output)

class SetVariableCommandTests(TestCase):
  def test_success_scenario(self):
    command = SetVariableCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', name = 'testName', value = 'testValue')   	
    desired_output = 'bgapi uuid_setvar 21516b8e-5a0b-485a-9e53-933e42947079 testName testValue\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

class SimplifyCommandTests(TestCase):
  def test_success_scenario(self):
    command = SimplifyCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', name = 'testName', value = 'testValue')   	
    desired_output = 'bgapi uuid_simplify 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)
 
class StartDebugMediaCommandTests(TestCase):
  def test_success_scenario(self):
    command = StartDebugMediaCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', option = 'testOption')   	
    desired_output = 'bgapi uuid_debug_media 21516b8e-5a0b-485a-9e53-933e42947079 testOption on\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)
 
class StartDisplaceCommandTests(TestCase):
  def test_success_scenario(self):
    command = StartDisplaceCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', path = 'testPath', limit = 450, mux='testMux')   	
    desired_output = 'bgapi uuid_displace 21516b8e-5a0b-485a-9e53-933e42947079 start testPath 450 mux\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)
 
class StopDebugMediaCommandTests(TestCase):
  def test_success_scenario(self):
    command = StopDebugMediaCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
    desired_output = 'bgapi uuid_debug_media 21516b8e-5a0b-485a-9e53-933e42947079 off\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

class StopDisplaceCommandTests(TestCase):
  def test_success_scenario(self):
    command = StopDisplaceCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')
    desired_output = 'bgapi uuid_displace 21516b8e-5a0b-485a-9e53-933e42947079 stop\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

class TransferCommandTests(TestCase):
  def test_success_scenario(self):
    command = TransferCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079', leg = 'legTest', extension = '384', dialplan = '44', context = 'contextTest')
    desired_output = 'bgapi uuid_transfer 21516b8e-5a0b-485a-9e53-933e42947079 legTest 384 44 contextTest\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)

class UnholdCommandTests(TestCase):
  def test_success_scenario(self):
    command = UnholdCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')   	
    desired_output = 'bgapi uuid_hold off 21516b8e-5a0b-485a-9e53-933e42947079\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)
 
class UnpauseCommandTests(TestCase):
  def test_success_scenario(self):
    command = UnpauseCommand(object(), '21516b8e-5a0b-485a-9e53-933e42947079')   	
    desired_output = 'bgapi pause 21516b8e-5a0b-485a-9e53-933e42947079 off\nJob-UUID: %s\n\n' % command.__job_uuid__
    self.assertTrue(str(command) == desired_output)
 


from setuptools import setup, find_packages

setup(
    name="freepy",
    version="0.9.8",
    packages=find_packages('src'),
    package_dir = {'':'src'},
    install_requires=['pykka == 1.2.0', 'llist == 0.4', 'twisted == 13.2.0'],
    include_package_data = True,
    author="Thomas Quintana",
    author_email="quintana.thomas@gmail.com",
    license="Apache License 2.0",
    url="https://github.com/thomasquintana/freepy",
    entry_points={
        'console_scripts': [
          'freepy-server = freepy.run:main',
        ]
    }
)




# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Thomas Quintana <quintana.thomas@gmail.com>

'''
In this module we implement a declarative finite state machine using method decorators.
'''
import types

class FiniteStateMachineError(Exception):
  '''
  A finite state machine exception.
  '''

  def __init__(self, message):
    self.message = message

  def __str__(self):
    return self.message

class Action(object):
  '''
  The Action decorator adds metadata to methods so they can be used by the finite
  state machine as actions that are executed upon entering or exiting a state.

  Arguments: state    - The state upon which the action will be executed.
             on_enter - If True the action is executed when entering the state.
             on_exit  - If True the action is executed when exiting the state.
  '''

  def __init__(self, *args, **kwargs):
    self.args = args
    self.kwargs = kwargs

  def __call__(self, function):
    # Append metadata to the function that is necessary
    # to build a transition lookup table by the finite
    # state machine.
    state = self.kwargs.get('state')
    if not state \
      or not type(state) == types.StringType \
      or len(state) == 0:
      raise FiniteStateMachineError('Please specify a valid action \
      state attribute.\n Possible values are strings.')
    else:
      function.__fsm_action_state__ = state
    if self.kwargs.has_key('on_enter'):
      on_enter = self.kwargs.get('on_enter')
      if type(on_enter) == types.BooleanType:
        function.__fsm_action_on_enter__ = on_enter
      else:
        raise TypeError('Please specify a valid action on_enter \
        attribute.\n Possible values are True or False.')
    else:
      function.__fsm_action_on_enter__ = True
    if self.kwargs.has_key('on_exit'):
      on_exit = self.kwargs.get('on_exit')
      if type(on_exit) == types.BooleanType:
        function.__fsm_action_on_exit__ = on_exit
      else:
        raise TypeError('Please specify a valid action on_exit \
        attribute.\n Possible values are True or False.')
    else:
      function.__fsm_action_on_exit__ = False
    function.__fsm_action__ = True
    return function

class Guard(object):
  '''
  The Guard decorator adds metadata to predicate methods so they can be used by 
  the finite state machine as guards protecting transitions into states.

  Arguments: state - The state upon which the guard will be executed.
  '''

  def __init__(self, *args, **kwargs):
    self.args = args
    self.kwargs = kwargs

  def __call__(self, function):
    # Append metadata to the function that is necessary
    # to build a transition lookup table by the finite
    # state machine.
    state = self.kwargs.get('state')
    if not state \
      or not type(state) == types.StringType \
      or len(state) == 0:
      raise TypeError('Please specify a valid guard state attribute.\n\
      Possible values are strings.')
    else:
      function.__fsm_guard_state__ = state
    function.__fsm_guard__ = True
    return function

class FiniteStateMachine(object):
  '''
  A finite state machine.
  '''
  def __init__(self, *args, **kwargs):
    super(FiniteStateMachine, self).__init__(*args, **kwargs)
    self.__fsm_transition_table__ = self.__create_lookup_table__()
    self.__state__ = self.initial_state

  def __create_lookup_table__(self):
    '''
    Creates a transition lookup table based on the possible transitions.
    '''
    actions = self.__get_actions__()
    guards = self.__get_guards__()
    states = self.__get_states__()
    state_map = self.__create_state_map__(actions, guards, states)
    # Create the lookup table.
    lookup_table = dict()
    for begin, end in self.transitions:
      transitions = lookup_table.get(begin)
      if not transitions:
        transitions = dict()
        lookup_table.update({ begin: transitions })
      transition = transitions.get(end)
      if not transition:
        transition = dict()
        transitions.update({ end: transition })
      if not transition.has_key('beginning_state'):
        transition.update({ 'beginning_state': state_map.get(begin) })
      if not transition.has_key('end_state'):
        transition.update({ 'end_state': state_map.get(end) })
    return lookup_table

  def __create_state_map__(self, actions, guards, states):
    '''
    Creates a map from states to actions and guards.

    Arguments: actions - The actions declared for this finite state machine.
               guards  - The guards declared for this finite state machine.
               state   - The possible states for this finite state machine.
    '''
    state_map = dict()
    # Make sure every state has an entry.
    for state in states:
      state_map.update({ state: dict() })
    # Attach actions to their states.
    for action in actions:
      state_name = action.__fsm_action_state__
      if state_name not in states:
        raise FiniteStateMachineError('A state named %s is not declared \
        in the transitions list.\n Please add %s to the list of possible \
        transitions or modify the @Action decorator on %s.' % (state_name, 
        state_name, action.__name__))
      state = state_map.get(state_name)
      on_enter = action.__fsm_action_on_enter__
      if not state.has_key('on_enter'):
        if on_enter:
          state.update({ 'on_enter': action })
      else:
        raise FiniteStateMachineError('The %s state can only have one \
        action declared for on_enter.' % (state_name))
      on_exit = action.__fsm_action_on_exit__
      if not state.has_key('on_exit'):
        if on_exit:
          state.update({ 'on_exit': action })
      else:
        raise FiniteStateMachineError('The %s state can only have one \
        action declared for on_exit.' % (state_name))
    # Attach guards to their states.
    for guard in guards:
      state_name = guard.__fsm_guard_state__
      if state_name not in states:
        raise FiniteStateMachineError('A state named %s is not declared \
        in the transitions list.\n Please add %s to the list of possible \
        transitions or modify the @Guard decorator on %s.' % (state_name,
        state_name, guard.__name__))
      state = state_map.get(state_name)
      if not state.has_key('guard'):
        state.update({ 'guard': guard })
      else:
        raise FiniteStateMachineError('The %s state can only have one guard \
        declared' % (state_name))
    return state_map

  def __get_actions__(self):
    '''
    Returns: All the actions declared for this finite state machine.
    '''
    return filter(
      lambda callable: hasattr(callable, '__fsm_action__'),
      self.__get_callables__()
    )

  def __get_callables__(self):
    '''
    Returns: All the methods for this object.
    '''
    actions = list()
    results = dir(self)
    for result in results:
      attr = getattr(self, result)
      if hasattr(attr, '__call__'):
        actions.append(attr)
    return actions

  def __get_guards__(self):
    '''
    Returns: All the guards declared for this finite state machine.
    '''
    return filter(
      lambda callable: hasattr(callable, '__fsm_guard__'),
      self.__get_callables__()
    )

  def __get_states__(self):
    '''
    Returns: The possible states based on the declared transitions.
    '''
    # Make sure the user declared a set of possible transitions.
    if not self.transitions:
      raise FiniteStateMachineError('Please specify a list of possible \
      transitions.\n Each entry in the list is a two-tuple where the \
      first value is the beginning state and the second value is the \
      end state.')
    states = list()
    for begin, end in self.transitions:
      if begin not in states:
        states.append(begin)
      if end not in states:
        states.append(end)
    return states

  def state(self):
    return self.__state__

  def transition(self, to = None, event = None):
    '''
    Transitions the finite state machine to a new state.

    Arguments: to - The desired end state.
               event - The event that caused the state change.
    '''
    # Make sure we are in a good state.
    transitions = self.__fsm_transition_table__.get(self.__state__)
    if not transitions:
      raise FiniteStateMachineError('The %s state is invalid, or we \
      have entered a terminal state.' % (self.__state__))
    # Try to find the desired transition.
    transition = transitions.get(to)
    if not transition:
      raise FiniteStateMachineError('The transition from %s to %s is \
      invalid.' % (self.__state__, to))
    # If there are any guards lets execute those now.
    if transition.get('end_state').has_key('guard'):
      allowed = transition.get('end_state').get('guard')()
      if not type(allowed) == types.BooleanType:
        raise FiniteStateMachineError('A guard must only return True \
        or False values.')
      if not allowed:
        raise FiniteStateMachineError('A guard declined the transition \
        from %s to %s.' % (self.__state__, to))
    # Try to execute the action associated with leaving the current state.
    if transition.get('beginning_state').has_key('on_exit'):
      transition.get('beginning_state').get('on_exit')(event)
    # Try to execute the action associated with entering the new state.
    if transition.get('end_state').has_key('on_enter'):
      transition.get('end_state').get('on_enter')(event)
    # Enter the new state and we're done.
    self.__state__ = to


# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Thomas Quintana <quintana.thomas@gmail.com>

from freepy.lib.core import *
from freepy.lib.services import *
from pykka import ActorRegistry
from unittest import TestCase

import logging, math, time

logging.basicConfig(
  format = '%(asctime)s %(levelname)s - %(name)s - %(message)s',
  level = logging.DEBUG
)

class InitializeTestSwitchletEvent(object):
  def __init__(self, timeout, skew):
    self.__timeout__ = timeout
    self.__skew__ = skew

  def get_skew(self):
    return self.__skew__

  def get_timeout(self):
    return self.__timeout__

class TimerServiceTestSwitchlet(Switchlet):
  def __init__(self, *args, **kwargs):
    super(TimerServiceTestSwitchlet, self).__init__(*args, **kwargs)
    self.__last_time__ = 0
    self.__misses__ = 0
    self.__timeout__ = 0
    self.__skew__ = 0

  def on_receive(self, message):
    message = message.get('content')
    if isinstance(message, InitializeTestSwitchletEvent):
      self.__timeout__ = message.get_timeout()
      self.__skew__ = message.get_skew()
    elif isinstance(message, TimeoutEvent):
      now = time.time()
      if self.__last_time__ > 0:
        skew = self.__timeout__ - math.trunc((now - self.__last_time__) * 1000)
        if skew > self.__skew__ or skew < -self.__skew__:
          self.__misses__ += 1
        print 'last time: %f, now: %f, skew: %ims' % (self.__last_time__, now, skew)
      else:
        print 'Initialized timer service test switchlet.'
      self.__last_time__ = now

class TimerServiceTests(TestCase):
  @classmethod
  def tearDownClass(cls):
    ActorRegistry.stop_all()

  def test_recurring_timer_one_second_interval(self):
    # Start the timer service.
    service = TimerService().start()
    # Start the seconds switchlet.
    consumer = TimerServiceTestSwitchlet().start()
    event = InitializeTestSwitchletEvent(1000, 100)
    consumer.tell({'content': event})
    # Tell the timer service to send timeout events to the seconds switchlet every second.
    command = ReceiveTimeoutCommand(consumer, 1000, recurring = True)
    service.tell({'content': command})
    # Wait for 1 minute worth of events to fire.
    time.sleep(60.0)
    # Tell the timer service to stop timeout events.
    command = StopTimeoutCommand(consumer)
    service.tell({'content': command})
    # Make the timer service did not miss any deadlines.
    print 'misses: %i' % consumer._actor.__misses__
    self.assertTrue( consumer._actor.__misses__ == 0)
    # cleanup.
    consumer.stop()
    service.stop()


# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Cristian Groza  <frontc18@gmail.com>
#
# Thomas Quintana <quintana.thomas@gmail.com>

from freepy.lib.commands import *
from freepy.lib.core import *
from freepy.lib.fsm import *
from freepy.lib.server import Event, RegisterJobObserverCommand, UnregisterJobObserverCommand

import logging
import urllib

class AbstractIVRMenu(FiniteStateMachine, Switchlet):
  initial_state = 'not ready'

  transitions = [
    ('not ready', 'greeting'),
    ('greeting', 'routing'),
    ('routing', 'bridging'),
    ('bridging', 'routing'),
    ('routing', 'terminated'),
    ('routing', 'saying'),
    ('saying', 'routing'),
    ('routing', 'playing'),
    ('playing', 'routing'),
    ('routing', 'transfering'),
    ('transfering', 'routing')
  ]

  def __init__(self, *args, **kwargs):
    super(Monitor, self).__init__(*args, **kwargs)
    self.__logger__ = logging.getLogger('ivr.menu.AbstractIVRMenu')
    self.__dispatcher__ = None
    


# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#
# Thomas Quintana <quintana.thomas@gmail.com>

# Cristian Groza <frontc18@gmail.com>

from uuid import uuid4

# Import the proper StringIO implementation.
try:
  from cStringIO import StringIO
except:
  from StringIO import StringIO

class BackgroundCommand(object):
  def __init__(self, *args, **kwargs):
    self.__sender__ = args[0]
    self.__job_uuid__ = uuid4().get_urn().split(':', 2)[2]
    if not self.__sender__:
      raise ValueError('The sender parameter must be a valid reference to a Pykka actor.')

  def get_job_uuid(self):
    return self.__job_uuid__

  def get_sender(self):
    return self.__sender__

class UUIDCommand(BackgroundCommand):
  def __init__(self, *args, **kwargs):
    super(UUIDCommand, self).__init__(*args, **kwargs)
    self.__uuid__ = args[1]
    if not self.__uuid__:
      raise ValueError('The value of uuid must be a valid UUID.')

  def get_uuid(self):
    return self.__uuid__

class ACLCheckCommand(BackgroundCommand):
  '''
  The ACLCheckCommand compares an ip to an ACL list.

  Arguments: sender - The freepy actor sending this command.
             ip - Internet Protocol address.
             list_name - ACL list name.
  '''
  def __init__(self, *args, **kwargs):
    super(ACLCheckCommand, self).__init__(*args, **kwargs)
    self.__ip__ = kwargs.get('ip')
    self.__list_name__ = kwargs.get('list_name')
    
    if not self.__ip__ :
      raise ValueError('The ip value %s is invalid' % self.__ip__)
    if not self.__list_name__ :
      raise ValueError('The list name value %s is invalid' % self.__list_name__)
    
  def get_ip(self):
    return self.__ip__

  def get_list_name(self):
    return self.__list_name__

  def __str__(self):
    return 'bgapi acl %s %s\nJob-UUID: %s\n\n' % (self.__ip__,
      self.__list_name__, self.__job_uuid__)

class AnswerCommand(UUIDCommand):
  '''
  The AnswerCommand answers a channel.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(AnswerCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_answer %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class BreakCommand(UUIDCommand):
  '''
  Break out of media being sent to a channel. For example, if an audio file is being played to a channel, 
  issuing uuid_break will discontinue the media and the call will move on in the dialplan, script, 
  or whatever is controlling the call.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              stop_all - boolean flag.*

  *If the stop_all flag is used then all audio files/prompts/etc. that are queued up to be played to the channel 
  will be removed, whereas without the all flag only the currently playing file will be discontinued. 
  '''
  def __init__(self, *args, **kwargs):
    super(BreakCommand, self).__init__(*args, **kwargs)
    self.__stop_all__ = kwargs.get('all', False)

  def stop_all(self):
    return self.__stop_all__

  def __str__(self):
    if not self.__stop_all__:
      return 'bgapi uuid_break %s\nJob-UUID: %s\n\n' % (self.__uuid__,
        self.__job_uuid__)
    else:
      return 'bgapi uuid_break %s all\nJob-UUID: %s\n\n' % (self.__uuid__,
        self.__job_uuid__)

class BridgeCommand(UUIDCommand):
  '''
  Bridge two call legs together. Bridge needs atleast any one leg to be answered. 
  
  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              other_uuid - universal unique identifier to Bridge.
  '''
  def __init__(self, *args, **kwargs):
    super(BridgeCommand, self).__init__(*args, **kwargs)
    self.__other_uuid__ = kwargs.get('other_uuid')
    if not self.__other_uuid__:
      raise ValueError('The value of other_uuid must be a valid UUID.')

  def get_other_uuid(self):
    return self.__other_uuid__

  def __str__(self):
    return 'bgapi uuid_bridge %s %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__other_uuid__, self.__job_uuid__)

class BroadcastCommand(UUIDCommand):
  '''
  Execute an arbitrary dialplan application on a specific <uuid>. 
  If a filename is specified then it is played into the channel(s).
  To execute an application use <app_name> and <app_args> syntax.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              leg - select which leg(s) to use [aleg|bleg|both].
              path - the path to the audio file to be played into the channels.*
              app_name - application name to be executed on the channel(s).*2
              app_args - arguments to the application(s) being executed on the channel(s).
  
  * Can provide path Or app_name, but not both. 
  '''
  def __init__(self, *args, **kwargs):
    super(BroadcastCommand, self).__init__(*args, **kwargs)
    self.__leg__ = kwargs.get('leg')
    if not self.__leg__ or not self.__leg__ == 'aleg' and \
      not self.__leg__ == 'bleg' and not self.__leg__ == 'both':
      raise ValueError('The leg value %s is invalid' % self.__leg__)
    self.__path__ = kwargs.get('path')
    self.__app_name__ = kwargs.get('app_name')
    self.__app_args__ = kwargs.get('app_args')
    if self.__path__ and self.__app_name__:
      raise RuntimeError('A broadcast command can specify either a path \
      or an app_name but not both.')

  def get_leg(self):
    return self.__leg__

  def get_path(self):
    return self.__path__

  def get_app_name(self):
    return self.__app_name__

  def get_app_args(self):
    return self.__app_args__

  def __str__(self):
    buffer = StringIO()
    buffer.write('bgapi uuid_broadcast %s ' % self.__uuid__)
    if self.__path__:
      buffer.write('%s' % self.__path__)
    else:
      buffer.write('%s' % self.__app_name__)
      if self.__app_args__:
        buffer.write('::%s' % self.__app_args__)
    buffer.write(' %s\nJob-UUID: %s\n\n' % (self.__leg__, self.__job_uuid__))
    try:
      return buffer.getvalue()
    finally:
      buffer.close()

class ChatCommand(UUIDCommand):
  '''
  Send a chat message. If the endpoint associated with the session 
  <uuid> has a receive_event handler, this message gets sent to that 
  session and is interpreted as an instant message. 
  
  Arguments:  sender - the freepy actor sending this command.
              uuid - universal unique identifier.
              text - message to be sent.
  '''
  def __init__(self, *args, **kwargs):
    super(ChatCommand, self).__init__(*args, **kwargs)
    self.__text__ = kwargs.get('text', '')

  def get_text(self):
    return self.__text__

  def __str__(self):
    return 'bgapi uuid_chat %s %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__text__, self.__job_uuid__)

class CheckUserGroupCommand(BackgroundCommand):
  '''
  Determine if a <user> is in a group. 

  Arguments: sender - the freepy actor sending this command.
             'user' - username.
             'domain' - domain name.
             'group_name' - group name.
  '''  
  def __init__(self, *args, **kwargs):
    super(CheckUserGroupCommand, self).__init__(*args, **kwargs)
    self.__user__ = kwargs.get('user')
    self.__domain__ = kwargs.get('domain')
    self.__group_name__ = kwargs.get('group_name')

  def get_domain(self):
    return self.__domain__

  def get_group_name(self):
    return self.__group_name__

  def get_user(self):
    return self.__user__

  def __str__(self):
    if not self.__domain__:
      return 'bgapi in_group %s %s\nJob-UUID: %s\n\n' % (self.__user__,
        self.__group_name__, self.__job_uuid__)
    else:
      return 'bgapi in_group %s@%s %s\nJob-UUID: %s\n\n' % (self.__user__,
        self.__domain__, self.__group_name__, self.__job_uuid__)

class DeflectCommand(UUIDCommand):
  '''
  Deflect an answered SIP call off of FreeSWITCH by sending the REFER method.
  Deflect waits for the final response from the far end to be reported. 
  It returns the sip fragment from that response as the text in the FreeSWITCH 
  response to uuid_deflect. If the far end reports the REFER was successful, 
  then FreeSWITCH will issue a bye on the channel. 
  
  Arguments:  sender - the freepy actor sending this command.
              uuid - universal unique identifier.
              url - SIP url.
  '''
  def __init__(self, *args, **kwargs):
    super(DeflectCommand, self).__init__(*args, **kwargs)
    self.__url__ = kwargs.get('url')

  def get_url(self):
    return self.__url__

  def __str__(self):
    return 'bgapi uuid_deflect %s %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__url__, self.__job_uuid__)

class DialedExtensionHupAllCommand(BackgroundCommand):
  '''
  Can be used to disconnect existing calls to an extension.

  Arguments: sender - The freepy actor sending this command.
             clearing - clearing type. 
             extension - extension number to be disconnected.
  '''  
  def __init__(self, *args, **kwargs):
    super(DialedExtensionHupAllCommand, self).__init__(*args, **kwargs)
    self.__clearing__ = kwargs.get('clearing')
    self.__extension__ = kwargs.get('extension')

  def get_clearing(self):
    return self.__clearing__

  def get_extension(self):
    return self.__extension__

  def __str__(self):
    return 'bgapi fsctl hupall %s dialed_ext %s\nJob-UUID: %s\n\n' % (self.__clearing__,
      self.__extension__, self.__job_uuid__)

class DisableMediaCommand(UUIDCommand):
  '''
  Reinvite FreeSWITCH out of the media path: 
  
  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(DisableMediaCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_media off %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class DisableVerboseEventsCommand(BackgroundCommand):
  '''
  Disable verbose events. Verbose events have every channel variable in every event 
  for a particular channel. Non-verbose events have only the pre-selected channel 
  variables in the event headers. 

  Arguments: sender - The freepy actor sending this command.
  '''  
  def __init__(self, *args, **kwargs):
    super(DisableVerboseEventsCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi fsctl verbose_events off\nJob-UUID: %s\n\n' % self.__job_uuid__

class DisplayCommand(UUIDCommand):
  '''
  Updates the display on a phone if the phone supports this. This works on some SIP 
  phones right now including Polycom and Snom. This command makes the phone re-negotiate 
  the codec. The SIP -> RTP Packet Size should be 0.020. If it is set to 0.030 on the SPA 
  series phones it causes a DTMF lag. When DTMF keys are pressed on the phone they are can 
  be seen on the fs_cli 4-6 seconds late. 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              display - screen display.
  '''
  def __init__(self, *args, **kwargs):
    super(DisplayCommand, self).__init__(*args, **kwargs)
    self.__display__ = kwargs.get('display')

    def get_display(self):
      return self.__display__

  def __str__(self):
    return 'bgapi uuid_display %s %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__display__, self.__job_uuid__)

class DomainExistsCommand(BackgroundCommand):
  '''
  Check if a domain exists. 

  Arguments: sender - The freepy actor sending this command.
             domain - domain to check.
  '''  
  def __init__(self, *args, **kwargs):
    super(DomainExistsCommand, self).__init__(*args, **kwargs)
    self.__domain__ = kwargs.get('domain')

  def get_domain(self):
    return self.__domain__

  def __str__(self):
    return 'bgapi domain_exists %s\nJob-UUID: %s\n\n' % (self.__domain__,
      self.__job_uuid__)

class DualTransferCommand(UUIDCommand):
  '''
  Transfer each leg of a call to different destinations. 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              extension_a - target extension for aleg
              dialplan_a - target dialplan for aleg
              context_a - target context for aleg
              extension_b - target extension for bleg
              dialplan_b - target dialplan for bleg
              context_b - target context for bleg
  '''
  def __init__(self, *args, **kwargs):
    super(DualTransferCommand, self).__init__(*args, **kwargs)
    self.__extension_a__ = kwargs.get('extension_a')
    self.__extension_b__ = kwargs.get('extension_b')
    self.__dialplan_a__ = kwargs.get('dialplan_a')
    self.__dialplan_b__ = kwargs.get('dialplan_b')
    self.__context_a__ = kwargs.get('context_a')
    self.__context_b__ = kwargs.get('context_b')
    if not self.__extension_a__ and not self.__extension_b__:
      raise RuntimeError('A dual transer command requires the extension_a \
        and extension_b parameters to be provided.')

  def get_extension_a(self):
    return self.__extension_a__

  def get_extension_b(self):
    return self.__extension_b__

  def get_dialplan_a(self):
    return self.__dialplan_a__

  def get_dialplan_b(self):
    return self.__dialplan_b__

  def get_context_a(self):
    return self.__context_a__

  def get_context_b(self):
    return self.__context_b__

  def __str__(self):
    buffer = StringIO()
    buffer.write(self.__extension_a__)
    if self.__dialplan_a__:
      buffer.write('/%s' % self.__dialplan_a__)
    if self.__context_a__:
      buffer.write('/%s' % self.__context_a__)
    destination_a = buffer.getvalue()
    buffer.seek(0)
    buffer.write(self.__extension_b__)
    if self.__dialplan_b__:
      buffer.write('/%s' % self.__dialplan_b__)
    if self.__context_b__:
      buffer.write('/%s' % self.__context_b__)
    destination_b = buffer.getvalue()
    buffer.close()
    return 'bgapi uuid_dual_transfer %s %s %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      destination_a, destination_b, self.__job_uuid__)

class DumpCommand(UUIDCommand):
  '''
  Dumps all variable values for a session. 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              format - variable values output format. Default output XML.
  '''
  def __init__(self, *args, **kwargs):
    super(DumpCommand, self).__init__(*args, **kwargs)
    self.__format__ = kwargs.get('format', 'XML')

  def get_format(self):
    return self.__format__

  def __str__(self):
    return 'bgapi uuid_dump %s %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__format__, self.__job_uuid__)

class EarlyOkayCommand(UUIDCommand):
  '''
  Stops the process of ignoring early media, i.e. if ignore_early_media=true 
  it stops ignoring early media and responds normally. 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(EarlyOkayCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_early_ok %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class EnableMediaCommand(UUIDCommand):
  '''
  Reinvite FreeSWITCH into the media path: 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(EnableMediaCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_media %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class EnableSessionHeartbeatCommand(UUIDCommand):
  '''
  Enable session Heartbeat.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              start_time - seconds in the future to start.
  '''
  def __init__(self, *args, **kwargs):
    super(EnableSessionHeartbeatCommand, self).__init__(*args, **kwargs)
    self.__start_time__ = kwargs.get('start_time') # Seconds in the future to start

  def get_start_time(self):
    return self.__start_time__

  def __str__(self):
    if not self.__start_time__:
      return 'bgapi uuid_session_heartbeat %s\nJob-UUID: %s\n\n' % (self.__uuid__,
        self.__job_uuid__)
    else:
      return 'bgapi uuid_session_heartbeat %s sched %i\nJob-UUID: %s\n\n' % (self.__uuid__,
        self.__start_time__, self.__job_uuid__)

class EnableVerboseEventsCommand(BackgroundCommand):
  '''
  Enable verbose events. Verbose events have every channel variable in every event 
  for a particular channel. Non-verbose events have only the pre-selected channel 
  variables in the event headers. 

  Arguments: sender - The freepy actor sending this command.
  '''  

  def __init__(self, *args, **kwargs):
    super(EnableVerboseEventsCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi fsctl verbose_events on\nJob-UUID: %s\n\n' % self.__job_uuid__

class FileManagerCommand(UUIDCommand):
  '''
  Manage the audio being played into a channel from a sound file

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              command - a parameter which controls what action should be taken*
              value - the value of the command**

  * Available Commands : 
  [speed]
  [volume]
  [pause]
  [stop]
  [truncate]
  [restart]
  [seek]
  ** Unit of measurement is milliseconds
  '''
  def __init__(self, *args, **kwargs):
    super(FileManagerCommand, self).__init__(*args, **kwargs)
    self.__command__ = kwargs.get('command')
    if not self.__command__ or not self.__command__ == 'speed' and \
      not self.__command__ == 'volume' and not self.__command__ == 'pause' and \
      not self.__command__ == 'stop' and not self.__command__ == 'truncate' and \
      not self.__command__ == 'restart' and not self.__command__ == 'seek':
      raise ValueError('The command parameter %s is invalid.' % self.__command__)
    self.__value__ = kwargs.get('value')

  def get_command(self):
    return self.__command__

  def get_value(self):
    return self.__value__

  def __str__(self):
    if self.__value__:
      return 'bgapi uuid_fileman %s %s:%s\nJob-UUID: %s\n\n' % (self.__uuid__,
        self.__command__, self.__value__, self.__job_uuid__)
    else:
      return 'bgapi uuid_fileman %s %s\nJob-UUID: %s\n\n' % (self.__uuid__,
        self.__command__, self.__job_uuid__)

class FlushDTMFCommand(UUIDCommand):
  '''
  Flush queued DTMF digits.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(FlushDTMFCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_flush_dtmf %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class GetAudioLevelCommand(UUIDCommand):
  '''
  Get the Audio Level 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(GetAudioLevelCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_audio %s start read level\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class GetBugListCommand(UUIDCommand):
  '''
  List the media bugs on channel.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(GetBugListCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_buglist %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class GetDefaultDTMFDurationCommand(BackgroundCommand):
  '''
  Gets the current value of default dtmf duration.

  Arguments: sender - The freepy actor sending this command.
  '''  
  def __init__(self, *args, **kwargs):
    super(GetDefaultDTMFDurationCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi fsctl default_dtmf_duration 0\nJob-UUID: %s\n\n' % self.__job_uuid__

class GetGlobalVariableCommand(BackgroundCommand):
  '''
  Gets the value of a global variable. 

  Arguments: sender - The freepy actor sending this command.
             name - the name of a global variable*

  * If the parameter is not provided then it gets all the global variables. 
  '''  
  def __init__(self, *args, **kwargs):
    super(GetGlobalVariableCommand, self).__init__(*args, **kwargs)
    self.__name__ = kwargs.get('name')
    if not self.__name__:
      raise ValueError('The name parameter is required.')

  def get_name(self):
    return self.__name__

  def __str__(self):
    return 'bgapi global_getvar %s\nJob-UUID: %s\n\n' % (self.__name__,
      self.__job_uuid__)

class GetMaxSessionsCommand(BackgroundCommand):
  '''
  Gets the value of the Maximum Sessions. 

  Arguments: sender - The freepy actor sending this command.
  '''  
  def __init__(self, *args, **kwargs):
    super(GetMaxSessionsCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi fsctl max_sessions\nJob-UUID: %s\n\n' % self.__job_uuid__

class GetMaximumDTMFDurationCommand(BackgroundCommand):
  '''
  Gets the current value of maximum dtmf duration.

  Arguments: sender - The freepy actor sending this command.
  '''  
  def __init__(self, *args, **kwargs):
    super(GetMaximumDTMFDurationCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi fsctl max_dtmf_duration 0\nJob-UUID: %s\n\n' % self.__job_uuid__

class GetMinimumDTMFDurationCommand(BackgroundCommand):
  '''
  Gets the current value of minimum dtmf duration.

  Arguments: sender - The freepy actor sending this command.
  '''  
  def __init__(self, *args, **kwargs):
    super(GetMinimumDTMFDurationCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi fsctl min_dtmf_duration 0\nJob-UUID: %s\n\n' % self.__job_uuid__

class GetSessionsPerSecondCommand(BackgroundCommand):
  '''
  Query the actual sessions-per-second. 

  Arguments: sender - The freepy actor sending this command.
  '''  
  def __init__(self, *args, **kwargs):
    super(GetSessionsPerSecondCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi fsctl last_sps\nJob-UUID: %s\n\n' % self.__job_uuid__

class GetVariableCommand(UUIDCommand):
  '''
  Get a variable from a channel. 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              name - the name of the variable to get from a channel
  '''
  def __init__(self, *args, **kwargs):
    super(GetVariableCommand, self).__init__(*args, **kwargs)
    self.__name__ = kwargs.get('name')
    if not self.__name__:
      raise ValueError('The name parameter is requied.')

  def get_name(self):
    return self.__name__

  def __str__(self):
    return 'bgapi uuid_getvar %s %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__name__, self.__job_uuid__)

class GetGroupCallBridgeStringCommand(BackgroundCommand):
  '''
  Returns the bridge string defined in a call group.

  Arguments: sender - The freepy actor sending this command.
             group - group name
             domain - domain name
             option - valid options [+F, +A, +E] *

  * +F
      will return the group members in a serial fashion (separated by |), 
    +A 
      will return them in a parallel fashion (separated by ,) 
    +E 
      will return them in a enterprise fashion (separated by :_:). 
  '''  

  def __init__(self, *args, **kwargs):
    super(GetGroupCallBridgeStringCommand, self).__init__(*args, **kwargs)
    self.__group__ = kwargs.get('group')
    self.__domain__ = kwargs.get('domain')
    self.__option__ = kwargs.get('option')
    if self.__option__ and not self.__option__ == '+F' and \
      not self.__option__ == '+A' and not self.__option__ == '+E':
      raise ValueError('The option parameter %s is invalid.' % self.__option__)

  def get_domain(self):
    return self.__domain__

  def get_group(self):
    return self.__group__

  def get_option(self):
    return self.__option__

  def __str__(self):
    if not self.__option__:
      return 'bgapi group_call %s@%s\nJob-UUID: %s\n\n' % (self.__group__,
        self.__domain__, self.__job_uuid__)
    else:
      return 'bgapi group_call %s@%s%s\nJob-UUID: %s\n\n' % (self.__group__,
        self.__domain__, self.__option__, self.__job_uuid__)

class HoldCommand(UUIDCommand):
  '''
  Place a call on hold. 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(HoldCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_hold %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class HupAllCommand(BackgroundCommand):
  '''
  Disconnect existing channels. 

  Arguments: sender - The freepy actor sending this command.
             cause - the reason.
             var_name - optional parameter variable name.
             var_value - optional parameter variable value. 
  '''  

  def __init__(self, *args, **kwargs):
    super(HupAllCommand, self).__init__(*args, **kwargs)
    self.__cause__ = kwargs.get('cause')
    self.__var_name__ = kwargs.get('var_name')
    self.__var_value__ = kwargs.get('var_value')

  def get_cause(self):
    return self.__cause__

  def get_variable_name(self):
    return self.__var_name__

  def get_variable_value(self):
    return self.__var_value__

  def __str__(self):
    if self.__var_name__ and self.__var_value__:
      return 'bgapi hupall %s %s %s\nJob-UUID: %s\n\n' % (self.__cause__,
        self.__var_name__, self.__var_value__, self.__job_uuid__)
    else:
      return 'bgapi hupall %s\nJob-UUID: %s\n\n' % (self.__cause__,
        self.__job_uuid__)

class KillCommand(UUIDCommand):
  '''
  Reset a specific <uuid> channel. 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              cause - reason for reset.
  '''
  def __init__(self, *args, **kwargs):
    super(KillCommand, self).__init__(*args, **kwargs)
    self.__cause__ = kwargs.get('cause')

  def get_cause(self):
    return self.__cause__

  def __str__(self):
    if not self.__cause__:
      return 'bgapi uuid_kill %s\nJob-UUID: %s\n\n' % (self.__uuid__,
        self.__job_uuid__)
    else:
      return 'bgapi uuid_kill %s %s\nJob-UUID: %s\n\n' % (self.__uuid__,
        self.__cause__, self.__job_uuid__)

class LimitCommand(UUIDCommand):
  '''
  Apply or change limit(s) on a specified uuid. 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              backend - The backend to use. 
              realm - Arbitrary name.
              resource - The resource on which to limit the number of calls. 
              max_calls - The maximum number of concurrent calls allowed to pass 
                          or a call rate in calls/sec. If not set or set to a negative 
                          value the limit application only acts as a counter.
              interval - calls per second.
              number - destination number
              dialplan - destination dialplan
              context - destination context
  '''
  def __init__(self, *args, **kwargs):
    super(LimitCommand, self).__init__(*args, **kwargs)
    self.__backend__ = kwargs.get('backend')
    self.__realm__ = kwargs.get('realm')
    self.__resource__ = kwargs.get('resource')
    self.__max_calls__ = kwargs.get('max_calls')
    self.__interval__ = kwargs.get('interval')
    self.__number__ = kwargs.get('number')
    self.__dialplan__ = kwargs.get('dialplan')
    self.__context__ = kwargs.get('context')

  def get_backend(self):
    return self.__backend__

  def get_realm(self):
    return self.__realm__

  def get_resource(self):
    return self.__resource__

  def get_max_calls(self):
    return self.__max_calls__

  def __str__(self):
    buffer = StringIO()
    buffer.write('bgapi uuid_limit %s %s %s %s' % (self.__uuid__,
      self.__backend__, self.__realm__, self.__resource__))
    if self.__max_calls__:
      buffer.write(' %i' % self.__max_calls__)
      if self.__interval__:
        buffer.write('/%i' % self.__interval__)
    if self.__number__:
      buffer.write(' %s' % self.__number__)
      if self.__dialplan__:
        buffer.write(' %s' % self.__dialplan__)
        if self.__context__:
          buffer.write(' %s' % self.__context__)
    buffer.write('\nJob-UUID: %s\n\n' % self.__job_uuid__)
    try:
      return buffer.getvalue()
    finally:
      buffer.close()

class LoadModuleCommand(BackgroundCommand):
  '''
  Load external module 

  Arguments: sender - The freepy actor sending this command.
             name - module name.
  '''  
  def __init__(self, *args, **kwargs):
    super(LoadModuleCommand, self).__init__(*args, **kwargs)
    self.__name__ = kwargs.get('name')

  def get_name(self):
    return self.__name__

  def __str__(self):
    return 'bgapi load %s\nJob-UUID: %s\n\n' % (self.__name__,
      self.__job_uuid__)

class OriginateCommand(BackgroundCommand):
  '''
  Originate a new call. 

  Arguments: sender - The freepy actor sending this command.
             url - URL you are calling.
             extension - call extension.
             app_name - *
             app_args - arguments to the app specified by app_name
             options - **
  
  * These are valid application names that can be used in this context :
    park, bridge, javascript/lua/perl, playback (remove mod_native_file), and many others. 
  ** The following are options that can be passed :
      group_confirm_key
      group_confirm_file
      forked_dial
      fail_on_single_reject
      ignore_early_media
      return_ring_ready
      originate_retries
      originate_retry_sleep_ms
      origination_caller_id_name
      origination_caller_id_number
      originate_timeout
      sip_auto_answer 
  '''  

  def __init__(self, *args, **kwargs):
    super(OriginateCommand, self).__init__(*args, **kwargs)
    self.__url__ = kwargs.get('url')
    self.__extension__ = kwargs.get('extension')
    self.__app_name__ = kwargs.get('app_name')
    self.__app_args__ = kwargs.get('app_args', [])
    if not isinstance(self.__app_args__, list):
      raise TypeError('The app_args parameter must be a list type.')
    self.__options__ = kwargs.get('options', [])
    if not isinstance(self.__options__, list):
      raise TypeError('The options parameter must be a list type.')
    if self.__extension__ and self.__app_name__:
      raise RuntimeError('An originate command can specify either an \
      extension or an app_name but not both.')

  def get_app_name(self):
    return self.__app_name__

  def get_app_args(self):
    return self.__app_args__

  def get_extension(self):
    return self.__extension__

  def get_options(self):
    return self.__options__

  def get_url(self):
    return self.__url__

  def __str__(self):
    buffer = StringIO()
    buffer.write('bgapi originate ')
    if self.__options__:
      buffer.write('{%s}' % ','.join(self.__options__))
    buffer.write('%s ' % self.__url__)
    if self.__extension__:
      buffer.write('%s' % self.__extension__)
    else:
      if self.__app_args__:
        buffer.write('\'&%s(%s)\'' % (self.__app_name__, ' '.join(self.__app_args__)))
      else:
        buffer.write('&%s()' % self.__app_name__)
    buffer.write('\nJob-UUID: %s\n\n' % self.__job_uuid__)
    try:
      return buffer.getvalue()
    finally:
      buffer.close()

class ParkCommand(UUIDCommand):
  '''
  Park call 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(ParkCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_park %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class PauseCommand(UUIDCommand):
  '''
  Pause <uuid> media 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(PauseCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi pause %s on\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class PauseSessionCreationCommand(BackgroundCommand):
  '''
  inbound or outbound may optionally be specified to pause just inbound or outbound 
  session creation, both paused if nothing specified. resume has similar behavior. 

  Arguments: sender - The freepy actor sending this command.
             direction - inbound or outbound or None paramater value.
  '''  

  def __init__(self, *args, **kwargs):
    super(PauseSessionCreationCommand, self).__init__(*args, **kwargs)
    self.__direction__ = kwargs.get('direction')

  def get_direction(self):
    return self.__direction__

  def __str__(self):
    if not self.__direction__:
      return 'bgapi fsctl pause\nJob-UUID: %s\n\n' % self.__job_uuid__
    else:
      return 'bgapi fsctl pause %s\nJob-UUID: %s\n\n' % (self.__direction__,
        self.__job_uuid__)

class PreAnswerCommand(UUIDCommand):
  '''
  Preanswer a channel. 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(PreAnswerCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_pre_answer %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class PreProcessCommand(UUIDCommand):
  '''
  Pre-process Channel 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(PreProcessCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_preprocess %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class ReceiveDTMFCommand(UUIDCommand):
  '''
  Receve DTMF digits to <uuid> set.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              digits - Use the character w for a .5 second delay and the character W for a 1 second delay.
              tone_duration - Default tone duration is 2000ms. 
  '''
  def __init__(self, *args, **kwargs):
    super(ReceiveDTMFCommand, self).__init__(*args, **kwargs)
    self.__digits__ = kwargs.get('digits')
    self.__duration__ = kwargs.get('tone_duration')

  def get_digits(self):
    return self.__digits__

  def get_tone_duration(self):
    return self.__duration__

  def __str__(self):
    if not self.__duration__:
      return 'bgapi uuid_recv_dtmf %s %s\nJob-UUID: %s\n\n' % (self.__uuid__,
        self.__digits__, self.__job_uuid__)
    else:
      return 'bgapi uuid_recv_dtmf %s %s@%s\nJob-UUID: %s\n\n' % (self.__uuid__,
        self.__digits__, self.__duration__, self.__job_uuid__)

class ReclaimMemoryCommand(BackgroundCommand):
  '''
  Reclaim Memory

  Arguments: sender - The freepy actor sending this command.
  '''  

  def __init__(self, *args, **kwargs):
    super(ReclaimMemoryCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi fsctl reclaim_mem\nJob-UUID: %s\n\n' % self.__job_uuid__

class RenegotiateMediaCommand(UUIDCommand):
  '''
  API command to tell a channel to send a re-invite with optional list of new codecs

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              codec - audio/video codec.
  '''
  def __init__(self, *args, **kwargs):
    super(RenegotiateMediaCommand, self).__init__(*args, **kwargs)
    self.__codec__ = kwargs.get('codec')

  def get_codec(self):
    return self.__codec__

  def __str__(self):
    return 'bgapi uuid_media_reneg %s =%s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__codec__, self.__job_uuid__)

class ResumeSessionCreationCommand(BackgroundCommand):
  '''
  inbound or outbound may optionally be specified to resume just inbound or outbound session creation,
  both paused if nothing specified.

  Arguments: sender - The freepy actor sending this command.
             direction - inbound or outbound or None paramater value.
  '''  

  def __init__(self, *args, **kwargs):
    super(ResumeSessionCreationCommand, self).__init__(*args, **kwargs)
    self.__direction__ = kwargs.get('direction')

  def get_direction(self):
    return self.__direction__

  def __str__(self):
    if not self.__direction__:
      return 'bgapi fsctl resume\nJob-UUID: %s\n\n' % self.__job_uuid__
    else:
      return 'bgapi fsctl resume %s\nJob-UUID: %s\n\n' % (self.__direction__,
        self.__job_uuid__)

class SendDTMFCommand(UUIDCommand):
  '''
  Send DTMF digits to <uuid> set.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              digits - Use the character w for a .5 second delay and the character W for a 1 second delay.
              tone_duration - Default tone duration is 2000ms. 
  '''
  def __init__(self, *args, **kwargs):
    super(SendDTMFCommand, self).__init__(*args, **kwargs)
    self.__digits__ = kwargs.get('digits')
    self.__duration__ = kwargs.get('tone_duration')

  def get_digits(self):
    return self.__digits__

  def get_tone_duration(self):
    return self.__duration__

  def __str__(self):
    if not self.__duration__:
      return 'bgapi uuid_send_dtmf %s %s\nJob-UUID: %s\n\n' % (self.__uuid__,
        self.__digits__, self.__job_uuid__)
    else:
      return 'bgapi uuid_send_dtmf %s %s@%i\nJob-UUID: %s\n\n' % (self.__uuid__,
        self.__digits__, self.__duration__, self.__job_uuid__)

class SendInfoCommand(UUIDCommand):
  '''
  Send info to the endpoint.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(SendInfoCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_send_info %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class SetAudioLevelCommand(UUIDCommand):
  '''
  Adjust the audio levels on a channel or mute (read/write) via a media bug.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              level - is in the range from -4 to 4, 0 being the default value. 
  '''
  def __init__(self, *args, **kwargs):
    super(SetAudioLevelCommand, self).__init__(*args, **kwargs)
    self.__audio_level__ = kwargs.get('level')
    if not self.__audio_level__ or self.__audio_level__ < -4.0 or \
      self.__audio_level__ > 4.0:
      raise ValueError('The level value %s is invalid.' % self.__audio_level__)

  def get_level(self):
    return self.__audio_level__

  def __str__(self):
    return 'bgapi uuid_audio %s start write level %f\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__audio_level__, self.__job_uuid__)

class SetDefaultDTMFDurationCommand(BackgroundCommand):
  '''
  Sets the default_dtmf_duration switch parameter. The number is in clock ticks (CT) where 
  CT / 8 = ms. The default_dtmf_duration specifies the DTMF duration to use on 
  originated DTMF events or on events that are received without a duration specified. 
  This value can be increased or lowered. This value is lower-bounded by min_dtmf_duration 
  and upper-bounded by max_dtmf_duration.

  Arguments: sender - The freepy actor sending this command.
             duration - paramter value.
  '''  

  def __init__(self, *args, **kwargs):
    super(SetDefaultDTMFDurationCommand, self).__init__(*args, **kwargs)
    self.__duration__ = kwargs.get('duration')

  def get_duration(self):
    return self.__duration__

  def __str__(self):
    return 'bgapi fsctl default_dtmf_duration %i\nJob-UUID: %s\n\n' % (self.__duration__,
      self.__job_uuid__)

class SetGlobalVariableCommand(BackgroundCommand):
  '''
  Sets the value of a global variable. 

  Arguments: sender - The freepy actor sending this command.
             name - name of global variable
             value - value of global variable
  '''  

  def __init__(self, *args, **kwargs):
    super(SetGlobalVariableCommand, self).__init__(*args, **kwargs)
    self.__name__ = kwargs.get('name')
    self.__value__ = kwargs.get('value')
    if not self.__name__ or not self.__value__:
      raise RuntimeError('The set global variable command requires both name \
      and value parameters.')

  def get_name(self):
    return self.__name__

  def get_value(self):
    return self.__value__

  def __str__(self):
    return 'bgapi global_setvar %s=%s\nJob-UUID: %s\n\n' % (self.__name__,
      self.__value__, self.__job_uuid__)

class SetMaximumDTMFDurationCommand(BackgroundCommand):
  '''
  Sets the max_dtmf_duration switch parameter. The number is in clock ticks (CT) where CT / 8 = ms. 
  The max_dtmf_duration caps the playout of a DTMF event at the specified duration. 
  Events exceeding this duration will be truncated to this duration. 
  You cannot configure a duration on a profile that exceeds this setting. 
  This setting can be lowered, but cannot exceed 192000 (the default). 
  This setting cannot be set lower than min_dtmf_duration.

  Arguments: sender - The freepy actor sending this command.
             duration - the maximum duration if a DTMF event. 
  '''  

  def __init__(self, *args, **kwargs):
    super(SetMaximumDTMFDurationCommand, self).__init__(*args, **kwargs)
    self.__duration__ = kwargs.get('duration')

  def get_duration(self):
    return self.__duration__

  def __str__(self):
    return 'bgapi fsctl max_dtmf_duration %i\nJob-UUID: %s\n\n' % (self.__duration__,
      self.__job_uuid__)

class SetMinimumDTMFDurationCommand(BackgroundCommand):
  '''
  Sets the min_dtmf_duration switch parameter to 100ms. The number is in clock ticks where clockticks / 8 = ms. 
  The min_dtmf_duration specifies the minimum DTMF duration to use on outgoing events. 
  Events shorter than this will be increased in duration to match min_dtmf_duration. 
  You cannot configure a DTMF duration on a profile that is less than this setting. 
  You may increase this value, but cannot set it lower than 400 (the default). 
  This value cannot exceed max_dtmf_duration.

  Arguments: sender - The freepy actor sending this command.
             duration - value of parameter.
  '''  

  def __init__(self, *args, **kwargs):
    super(SetMinimumDTMFDurationCommand, self).__init__(*args, **kwargs)
    self.__duration__ = kwargs.get('duration')

  def get_duration(self):
    return self.__duration__

  def __str__(self):
    return 'bgapi fsctl min_dtmf_duration %i\nJob-UUID: %s\n\n' % (self.__duration__,
      self.__job_uuid__)

class SetMultipleVariableCommand(UUIDCommand):
  '''
  Set multiple vars on a channel. 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              variables - a dictionary list of key/value pairs to set.
  '''
  def __init__(self, *args, **kwargs):
    super(SetMultipleVariableCommand, self).__init__(*args, **kwargs)
    self.__variables__ = kwargs.get('variables')
    if not isinstance(self.__variables__, dict):
      raise TypeError('The variables parameter must be of type dict.')
    variable_list = list()
    for key, value in self.__variables__.iteritems():
      variable_list.append('%s=%s' % (key, value))
    self.__variables_string__ = ';'.join(variable_list)

  def __str__(self):
    return 'bgapi uuid_setvar_multi %s %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__variables_string__, self.__job_uuid__)

class SetSessionsPerSecondCommand(BackgroundCommand):
  '''
  This changes the sessions-per-second limit as initially set in switch.conf 

  Arguments: sender - The freepy actor sending this command.
             sessions_per_second - value for paramater.
  '''  

  def __init__(self, *args, **kwargs):
    super(SetSessionsPerSecondCommand, self).__init__(*args, **kwargs)
    self.__sessions_per_second__ = kwargs.get('sessions_per_second')

  def get_sessions_per_second(self):
    return self.__sessions_per_second__

  def __str__(self):
    return 'bgapi fsctl sps %i\nJob-UUID: %s\n\n' % (self.__sessions_per_second__,
      self.__job_uuid__)

class SetVariableCommand(UUIDCommand):
  '''
  Set a variable on a channel.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              name - name of the variable.
              value - value of the variable.
  '''
  def __init__(self, *args, **kwargs):
    super(SetVariableCommand, self).__init__(*args, **kwargs)
    self.__name__ = kwargs.get('name')
    self.__value__ = kwargs.get('value')
    if not self.__name__ or not self.__value__:
      raise RuntimeError('The set variable command requires both name \
      and value parameters.')

  def get_name(self):
    return self.__name__

  def get_value(self):
    return self.__value__

  def __str__(self):
    return 'bgapi uuid_setvar %s %s %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__name__, self.__value__, self.__job_uuid__)

class ShutdownCommand(BackgroundCommand):
  '''
  Stop the FreeSWITCH program. 

  Arguments: sender - The freepy actor sending this command.
             option - paramater value [cancel, elegant, asap, restart]*
             
  * available options: 
    cancel - discontinue a previous shutdown request.
    elegant - wait for all traffic to stop; do not prevent new traffic.
    asap - wait for all traffic to stop; do not allow new traffic.
    restart - restart FreeSWITCH immediately following the shutdown. 
  '''  

  def __init__(self, *args, **kwargs):
    super(ShutdownCommand, self).__init__(*args, **kwargs)
    self.__option__ = kwargs.get('option')
    if self.__option__ and not self.__option__ == 'cancel' and \
      not self.__option__ == 'elegant' and not self.__option__ == 'asap' and \
      not self.__option__ == 'restart':
      raise ValueError('The option %s is an invalid option.' % self.__option__)

  def get_option(self):
    return self.__option__

  def __str__(self):
    if not self.__option__:
      return 'bgapi fsctl shutdown\nJob-UUID: %s\n\n' % self.__job_uuid__
    else:
      return 'bgapi fsctl shutdown %s\nJob-UUID: %s\n\n' % (self.__option__,
        self.__job_uuid__)

class SimplifyCommand(UUIDCommand):
  '''
  This command directs FreeSWITCH to remove itself from the SIP signaling path if it can safely do so 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(SimplifyCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_simplify %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class StartDebugMediaCommand(UUIDCommand):
  '''
  Start the Debug Media.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              option - [read|write|both|vread|vwrite|vboth]
  '''
  def __init__(self, *args, **kwargs):
    super(StartDebugMediaCommand, self).__init__(*args, **kwargs)
    self.__option__ = kwargs.get('option')

  def get_option(self):
    return self.__option__

  def __str__(self):
    return 'bgapi uuid_debug_media %s %s on\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__option__, self.__job_uuid__)

class StartDisplaceCommand(UUIDCommand):
  '''
  Displace the audio for the target <uuid> with the specified audio <path>. 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              path - path to an audio source (wav, shout, etc...) 
              limit - number of seconds before terminating the displacement
              mux - cause the original audio to be mixed together with 'file', 
              i.e. you can still converse with the other party while the file is playing
  '''
  def __init__(self, *args, **kwargs):
    super(StartDisplaceCommand, self).__init__(*args, **kwargs)
    self.__path__ = kwargs.get('path')
    self.__limit__ = kwargs.get('limit')
    self.__mux__ = kwargs.get('mux')

  def get_limit(self):
    return self.__limit__

  def get_mux(self):
    return self.__mux__

  def get_path(self):
    return self.__path__

  def __str__(self):
    buffer = StringIO()
    buffer.write('bgapi uuid_displace %s start %s' % (self.__uuid__,
      self.__path__))
    if self.__limit__:
      buffer.write(' %i' % self.__limit__)
    if self.__mux__:
      buffer.write(' mux')
    buffer.write('\nJob-UUID: %s\n\n' % self.__job_uuid__)
    try:
      return buffer.getvalue()
    finally:
      buffer.close()

class StatusCommand(BackgroundCommand):
  '''
  Show current status 

  Arguments: sender - The freepy actor sending this command.
  '''  
  def __init__(self, *args, **kwargs):
    super(StatusCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi status\nJob-UUID: %s\n\n' % self.__job_uuid__

class StopDebugMediaCommand(UUIDCommand):
  '''
  Used to stop the debug media.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(StopDebugMediaCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_debug_media %s off\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class StopDisplaceCommand(UUIDCommand):
  '''
  Stop displacing the audio for the target <uuid>

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(StopDisplaceCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_displace %s stop\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class SyncClockCommand(BackgroundCommand):
  '''
  FreeSWITCH will not trust the system time. It gets one sample of system time when it first starts and 
  uses the monotonic clock from there. You can sync it back to real time with "fsctl sync_clock"
  Note: 'fsctl sync_clock' immediately takes effect - which can affect the times on your CDRs. 
  You can end up undrebilling/overbilling, or even calls hungup before they originated. 
  e.g. if FS clock is off by 1 month, then your CDRs are going to show calls that lasted for 1 month!
  'fsctl sync_clock_when_idle' is much safer, which does the same sync but doesn't take effect until 
  there are 0 channels in use. That way it doesn't affect any CDRs. 

  Arguments: sender - The freepy actor sending this command.
  '''  

  def __init__(self, *args, **kwargs):
    super(SyncClockCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi fsctl sync_clock\nJob-UUID: %s\n\n' % self.__job_uuid__

class SyncClockWhenIdleCommand(BackgroundCommand):
  '''
  You can sync the clock but have it not do it till there are 0 calls (r:2094f2d3) 

  Arguments: sender - The freepy actor sending this command.
  '''  
  def __init__(self, *args, **kwargs):
    super(SyncClockWhenIdleCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi fsctl sync_clock_when_idle\nJob-UUID: %s\n\n' % self.__job_uuid__

class TransferCommand(UUIDCommand):
  '''
  Transfers an existing call to a specific extension within a <dialplan> and <context>.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              leg - bleg or both.
              extension - destination extension.  
              dialplan - destination dialplan.*   
              context - destination context.
  * Dialplan may be "xml" or "directory".
  '''
  def __init__(self, *args, **kwargs):
    super(TransferCommand, self).__init__(*args, **kwargs)
    self.__leg__ = kwargs.get('leg')
    self.__extension__ = kwargs.get('extension')
    self.__dialplan__ = kwargs.get('dialplan')
    self.__context__ = kwargs.get('context')

  def get_context(self):
    return self.__context__

  def get_dialplan(self):
    return self.__dialplan__

  def get_extension(self):
    return self.__extension__

  def get_leg(self):
    return self.__leg__

  def __str__(self):
    buffer = StringIO()
    buffer.write('bgapi uuid_transfer %s' % self.__uuid__)
    if self.__leg__:
      buffer.write(' %s' % self.__leg__)
    buffer.write(' %s' % self.__extension__)
    if self.__dialplan__:
      buffer.write(' %s' % self.__dialplan__)
    if self.__context__:
      buffer.write(' %s' % self.__context__)
    buffer.write('\nJob-UUID: %s\n\n' % self.__job_uuid__)
    try:
      return buffer.getvalue()
    finally:
      buffer.close()

class UnholdCommand(UUIDCommand):
  '''
  Take a call off hold.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(UnholdCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi uuid_hold off %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class UnloadModuleCommand(BackgroundCommand):
  '''
  Unload external module. 

  Arguments: sender - The freepy actor sending this command.
             name - name of external module to unload.
             force - force unload. Default set to False.
  '''  

  def __init__(self, *args, **kwargs):
    super(UnloadModuleCommand, self).__init__(*args, **kwargs)
    self.__name__ = kwargs.get('name')
    self.__force__ = kwargs.get('force', False)

  def get_name(self):
    return self.__name__

  def get_force(self):
    return self.__force__

  def __str__(self):
    if not self.__force__:
      return 'bgapi unload %s\nJob-UUID: %s\n\n' % (self.__name__,
        self.__job_uuid__)
    else:
      return 'bgapi unload -f %s\nJob-UUID: %s\n\n' % (self.__name__,
        self.__job_uuid__)

class UnpauseCommand(UUIDCommand):
  '''
  UnPause <uuid> media.

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
  '''
  def __init__(self, *args, **kwargs):
    super(UnpauseCommand, self).__init__(*args, **kwargs)

  def __str__(self):
    return 'bgapi pause %s off\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__job_uuid__)

class UnsetVariableCommand(UUIDCommand):
  '''
  UnSet a variable on a channel. 

  Arguments:  sender - The freepy actor sending this command.
              uuid - universal unique identifier.
              name - variable to be unset.
  '''
  def __init__(self, *args, **kwargs):
    super(UnsetVariableCommand, self).__init__(*args, **kwargs)
    self.__name__ = kwargs.get('name')
    if not self.__name__:
      raise RuntimeError('The unset variable commands requires the name \
      parameter.')

  def get_name(self):
    return self.__name__

  def __str__(self):
    return 'bgapi uuid_setvar %s %s\nJob-UUID: %s\n\n' % (self.__uuid__,
      self.__name__, self.__job_uuid__)
